@@ -1,15 +1,26 @@
 import math
 from datetime import datetime, timedelta
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
 
 import altair as alt
 import pandas as pd
+import pytz
 from pandas.core.groupby import DataFrameGroupBy
-from sqlalchemy import Column, DateTime, Float, ForeignKey, Integer, Interval
+from sqlalchemy import (
+    Column,
+    DateTime,
+    Float,
+    ForeignKey,
+    Integer,
+    Interval,
+    and_,
+    func,
+)
 from sqlalchemy.ext.declarative import declared_attr, has_inherited_table
 from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
 from sqlalchemy.orm import Session, backref, relationship
 from sqlalchemy.schema import UniqueConstraint
+from sqlalchemy.sql.elements import BinaryExpression
 
 import timely_beliefs.utils as tb_utils
 from timely_beliefs.beliefs import probabilistic_utils
@@ -285,30 +296,47 @@ class TimedBeliefDBMixin(TimedBelief):
         return cls.search_session(*args, **kwargs)
 
     @classmethod
-    def search_session(
+    def search_session(  # noqa: C901  # todo: remove after removing deprecated arguments
         cls,
         session: Session,
-        sensor: SensorDBMixin,
+        sensor: Union[SensorDBMixin, int],
+        sensor_class: Optional[Type[SensorDBMixin]] = DBSensor,
         event_starts_after: Optional[datetime] = None,
         event_ends_before: Optional[datetime] = None,
         beliefs_after: Optional[datetime] = None,
         beliefs_before: Optional[datetime] = None,
+        horizons_at_least: Optional[timedelta] = None,
+        horizons_at_most: Optional[timedelta] = None,
         event_before: Optional[datetime] = None,  # deprecated
         event_not_before: Optional[datetime] = None,  # deprecated
         belief_before: Optional[datetime] = None,  # deprecated
         belief_not_before: Optional[datetime] = None,  # deprecated
         source: Optional[Union[BeliefSource, List[BeliefSource]]] = None,
+        most_recent_beliefs_only: bool = False,
+        most_recent_events_only: bool = False,
+        most_recent_only: bool = False,  # deprecated
+        place_beliefs_in_sensor_timezone: bool = True,
+        place_events_in_sensor_timezone: bool = True,
+        custom_filter_criteria: List[BinaryExpression] = None,
     ) -> "BeliefsDataFrame":
         """Search a database session for beliefs about sensor events.
 
         The optional arguments represent optional filters.
         :param session: the database session to use
-        :param sensor: sensor to which the beliefs pertain
+        :param sensor: sensor to which the beliefs pertain, or its unique sensor id
+        :param sensor_class: optionally pass the sensor (sub)class explicitly (only needed if you pass a sensor id instead of a sensor, and your sensor class is not DBSensor); the class should be mapped to a database table
         :param event_starts_after: only return beliefs about events that start after this datetime (inclusive)
         :param event_ends_before: only return beliefs about events that end before this datetime (inclusive)
         :param beliefs_after: only return beliefs formed after this datetime (inclusive)
         :param beliefs_before: only return beliefs formed before this datetime (inclusive)
+        :param horizons_at_least: only return beliefs with a belief horizon equal or greater than this timedelta (for example, use timedelta(0) to get ante knowledge time beliefs)
+        :param horizons_at_most: only return beliefs with a belief horizon equal or less than this timedelta (for example, use timedelta(0) to get post knowledge time beliefs)
         :param source: only return beliefs formed by the given source or list of sources
+        :param most_recent_beliefs_only: only return the most recent beliefs for each event from each source (minimum belief horizon)
+        :param most_recent_events_only: only return (post knowledge time) beliefs for the most recent event (maximum event start)
+        :param place_beliefs_in_sensor_timezone: if True (the default), belief times are converted to the timezone of the sensor
+        :param place_events_in_sensor_timezone: if True (the default), event starts are converted to the timezone of the sensor
+        :param custom_filter_criteria: optionally pass additional filters, such as ones that rely on subclasses
         :returns: a multi-index DataFrame with all relevant beliefs
         """
 
@@ -344,6 +372,14 @@ class TimedBeliefDBMixin(TimedBelief):
             beliefs_after,
             required_argument=False,
         )
+        # todo: deprecate the 'most_recent_only' argument in favor of 'most_recent_beliefs_only' (announced v1.7.0)
+        most_recent_beliefs_only = tb_utils.replace_deprecated_argument(
+            "most_recent_only",
+            most_recent_only,
+            "most_recent_beliefs_only",
+            most_recent_beliefs_only,
+            required_argument=False,
+        )
 
         # Check for timezone-aware datetime input
         if event_starts_after is not None:
@@ -363,25 +399,29 @@ class TimedBeliefDBMixin(TimedBelief):
                 beliefs_before, "belief_before"
             )
 
-        # Query sensor for relevant timing properties
-        event_resolution, knowledge_horizon_fnc, knowledge_horizon_par = (
-            session.query(
-                sensor.__class__.event_resolution,
-                sensor.__class__.knowledge_horizon_fnc,
-                sensor.__class__.knowledge_horizon_par,
+        # Query sensor, required for its timing properties
+        if isinstance(sensor, int):
+            # Check for proper sensor class
+            if not issubclass(sensor_class, SensorDBMixin):
+                raise ValueError(
+                    f"sensor {sensor} is a {type(sensor)}, which is not a subclass of {SensorDBMixin}"
+                )
+            sensor = (
+                session.query(sensor_class)
+                .filter(sensor_class.id == sensor)
+                .one_or_none()
             )
-            .filter(sensor.__class__.id == sensor.id)
-            .one_or_none()
-        )
+            if sensor is None:
+                raise ValueError("No such sensor")
 
         # Get bounds on the knowledge horizon (so we can already roughly filter by belief time)
         (
             knowledge_horizon_min,
             knowledge_horizon_max,
         ) = sensor_utils.eval_verified_knowledge_horizon_fnc(
-            knowledge_horizon_fnc,
-            knowledge_horizon_par,
-            event_resolution=event_resolution,
+            sensor.knowledge_horizon_fnc,
+            sensor.knowledge_horizon_par,
+            event_resolution=sensor.event_resolution,
             get_bounds=True,
         )
 
@@ -392,7 +432,7 @@ class TimedBeliefDBMixin(TimedBelief):
         if event_starts_after is not None:
             q = q.filter(cls.event_start >= event_starts_after)
         if event_ends_before is not None:
-            q = q.filter(cls.event_start + event_resolution <= event_ends_before)
+            q = q.filter(cls.event_start + sensor.event_resolution <= event_ends_before)
 
         # Apply rough belief time filter
         if beliefs_after is not None:
@@ -406,6 +446,12 @@ class TimedBeliefDBMixin(TimedBelief):
                 <= beliefs_before + cls.belief_horizon + knowledge_horizon_max
             )
 
+        # Apply belief horizon filter
+        if horizons_at_least is not None:
+            q = q.filter(cls.belief_horizon >= horizons_at_least)
+        if horizons_at_most is not None:
+            q = q.filter(cls.belief_horizon <= horizons_at_most)
+
         # Apply source filter
         if source is []:
             return BeliefsDataFrame(sensor=sensor, beliefs=[])
@@ -414,6 +460,51 @@ class TimedBeliefDBMixin(TimedBelief):
             source_cls = sources[0].__class__
             q = q.join(source_cls).filter(cls.source_id.in_([s.id for s in sources]))
 
+        # Apply most recent beliefs filter
+        if most_recent_beliefs_only:
+            subq = (
+                session.query(
+                    cls.event_start,
+                    cls.source_id,
+                    func.min(cls.belief_horizon).label("most_recent_belief_horizon"),
+                )
+                .filter(cls.sensor_id == sensor.id)
+                .group_by(cls.event_start, cls.source_id)
+                .subquery()
+            )
+            q = q.join(
+                subq,
+                and_(
+                    cls.event_start == subq.c.event_start,
+                    cls.source_id == subq.c.source_id,
+                    cls.belief_horizon == subq.c.most_recent_belief_horizon,
+                ),
+            )
+
+        # Apply most recent events filter
+        if most_recent_events_only:
+            subq_most_recent_events = (
+                session.query(
+                    cls.source_id,
+                    func.max(cls.event_start).label("most_recent_event_start"),
+                )
+                .filter(cls.sensor_id == sensor.id)
+                .group_by(cls.source_id)
+                .subquery()
+            )
+            q = q.join(
+                subq_most_recent_events,
+                and_(
+                    cls.source_id == subq_most_recent_events.c.source_id,
+                    cls.event_start
+                    == subq_most_recent_events.c.most_recent_event_start,
+                ),
+            )
+
+        # Apply custom filter criteria
+        if custom_filter_criteria is not None:
+            q = q.filter(*custom_filter_criteria)
+
         # Build our DataFrame of beliefs
         df = BeliefsDataFrame(sensor=sensor, beliefs=q.all())
 
@@ -423,6 +514,12 @@ class TimedBeliefDBMixin(TimedBelief):
         if beliefs_before is not None:
             df = df[df.index.get_level_values("belief_time") <= beliefs_before]
 
+        # Convert timezone of beliefs and events to sensor timezone
+        if place_beliefs_in_sensor_timezone:
+            df = df.convert_timezone_of_belief_timing_index(sensor.timezone)
+        if place_events_in_sensor_timezone:
+            df = df.convert_timezone_of_event_timing_index(sensor.timezone)
+
         return df
 
 
@@ -649,10 +746,12 @@ class BeliefsDataFrame(pd.DataFrame):
 
             # Check for different sources with the same name
             unique_sources = set(belief.source for belief in beliefs)
-            unique_source_names = set(source.name for source in unique_sources)
-            if len(unique_source_names) != len(unique_sources):
+            unique_source_string_representations = set(
+                str(source) for source in unique_sources
+            )
+            if len(unique_source_string_representations) != len(unique_sources):
                 raise ValueError(
-                    "Source names must be unique. Cannot initialise BeliefsDataFrame given the following unique sources:\n%s"
+                    "String representations of sources must be unique. Cannot initialise BeliefsDataFrame given the following unique sources:\n%s"
                     % unique_sources
                 )
 
@@ -802,9 +901,46 @@ class BeliefsDataFrame(pd.DataFrame):
             self, "belief_horizon", self.belief_times
         )
 
+    def convert_index_from_event_end_to_start(self) -> "BeliefsDataFrame":
+        return tb_utils.replace_multi_index_level(self, "event_end", self.event_starts)
+
     def convert_index_from_event_start_to_end(self) -> "BeliefsDataFrame":
         return tb_utils.replace_multi_index_level(self, "event_start", self.event_ends)
 
+    def convert_timezone_of_belief_timing_index(
+        self, timezone: Union[str, pytz.timezone]
+    ) -> "BeliefsDataFrame":
+        if "belief_horizon" in self.index.names:
+            return self  # timedeltas don't have timezones
+        elif "belief_time" in self.index.names:
+            return tb_utils.replace_multi_index_level(
+                self,
+                "belief_time",
+                pd.to_datetime(self.belief_times, utc=True).tz_convert(timezone),
+            )
+        else:
+            raise ValueError(
+                "Missing level 'belief_horizon' or 'belief_time' in index."
+            )
+
+    def convert_timezone_of_event_timing_index(
+        self, timezone: Union[str, pytz.timezone]
+    ) -> "BeliefsDataFrame":
+        if "event_end" in self.index.names:
+            return tb_utils.replace_multi_index_level(
+                self,
+                "event_end",
+                pd.to_datetime(self.event_ends, utc=True).tz_convert(timezone),
+            )
+        elif "event_start" in self.index.names:
+            return tb_utils.replace_multi_index_level(
+                self,
+                "event_start",
+                pd.to_datetime(self.event_starts, utc=True).tz_convert(timezone),
+            )
+        else:
+            raise ValueError("Missing level 'event_start' or 'event_end' in index.")
+
     def drop_belief_time_or_horizon_index_level(self) -> "BeliefsDataFrame":
         return self.droplevel(
             "belief_horizon" if "belief_horizon" in self.index.names else "belief_time"
@@ -858,15 +994,25 @@ class BeliefsDataFrame(pd.DataFrame):
 
     @property
     def event_starts(self) -> pd.DatetimeIndex:
-        return pd.DatetimeIndex(self.index.get_level_values("event_start"))
+        if "event_start" in self.index.names:
+            return pd.DatetimeIndex(self.index.get_level_values("event_start"))
+        else:
+            return pd.DatetimeIndex(
+                self.event_ends.to_series(name="event_start").apply(
+                    lambda event_end: event_end - self.event_resolution
+                )
+            )
 
     @property
     def event_ends(self) -> pd.DatetimeIndex:
-        return pd.DatetimeIndex(
-            self.event_starts.to_series(name="event_end").apply(
-                lambda event_start: event_start + self.event_resolution
+        if "event_end" in self.index.names:
+            return pd.DatetimeIndex(self.index.get_level_values("event_end"))
+        else:
+            return pd.DatetimeIndex(
+                self.event_starts.to_series(name="event_end").apply(
+                    lambda event_start: event_start + self.event_resolution
+                )
             )
-        )
 
     @property
     def sources(self) -> pd.Index:
